// PPgauge class

#include "../common/connector.h"
#include "../common/packet.h"
#include "../common/timer-handler.h"
#include "../link/delay.h"
#include "../common/agent.h"
#include "tcp-st.h"

//
// Declarations
//

struct A_list {
	StTcpSink*	agent;
	int		port;
	int		idle;
	int		syn;
	A_list*	next;
};

class PPgauge;

class GaugeTimer : public TimerHandler {
public:
	GaugeTimer(PPgauge* g) : TimerHandler() { g_ = g; }
protected:
	void expire(Event *e);
	PPgauge* g_;
};

class PPgauge : public Connector {
public:
	PPgauge();
	int	command(int argc, const char*const* argv);
	void recv(Packet* p, Handler* h);
	void timeout();
protected:
	Tcl_Channel	channel_;
	GaugeTimer	idle_timer_;
	double 		timeout_;
	// Packet Pair measurement
	int				pp_def_wnd_;
	TracedInt		pp_wnd_;
	int				pp_cnt_;
	int				pp_bytes_;
	double			pp_start_;
	TracedDouble	pp_throughput_;
	void 			pp_calc(Packet* p);
	// New Window calculation
	double	bw_avail();
	double	bw_equal(double bw, int shares);
	double	bw2win(double bw, int seg_size);
	int 	max_rwnd();
	double	max_share();
	TracedDouble	c_wnd_;
	// Tracing
	void trace(TracedVar*);
	// Pointer to link object, so we can get bw & delay
	LinkDelay* link_;
	//Agent list - must choose which functions to keep in final version
	A_list* 	a_list_;
	TracedInt	a_cnt_;
	void a_insert(StTcpSink* a);
	void a_remove(StTcpSink* a);
	void a_remove(int port);
	void a_update_absolute();
	void a_update_delta();
	void a_print();
	int  a_notIdle_cnt();
};

//
// OTCL linkage
//

static class PPgaugeClass : public TclClass {
public:
	PPgaugeClass() : TclClass("Connector/PPgauge") {}
	TclObject* create(int, const char*const*) {
		return (new PPgauge());
	}
} class_PPgauge;

//
// Implementation
//

void GaugeTimer::expire(Event*) {
	g_->timeout();
}

PPgauge::PPgauge() :
	Connector(),
	idle_timer_(this),
	timeout_(0.1),
	channel_(0),
	pp_def_wnd_(50),
	pp_wnd_(50),
	pp_cnt_(0),
	pp_bytes_(0),
	pp_start_(0),
	pp_throughput_(0),
	a_list_(0),
	a_cnt_(0),
	c_wnd_(2.0)
{
	bind("pp_throughput_", &pp_throughput_);
	bind("pp_wnd_", &pp_wnd_);
	bind("c_wnd_", &c_wnd_);
	bind("timeout_", &timeout_);
	bind("a_cnt_", &a_cnt_);
	idle_timer_.resched(timeout_);
}

void PPgauge::timeout() {
	double now = Scheduler::instance().clock();
	pp_throughput_ = (double)(pp_bytes_)/(now - pp_start_);
	pp_start_ = now;
	pp_bytes_ = 0;
	pp_cnt_ = 0;
	idle_timer_.resched(timeout_);
	a_update_delta();
	A_list* pos = a_list_;
	while(pos) {
		pos->idle = 1;
		if(pos->syn==1) pos->idle = 0;
		pos = pos->next;
	}
}

// Calculate incoming throughput using windowed Packet Pair

void PPgauge::pp_calc(Packet *p) {
	double now = Scheduler::instance().clock();
	hdr_cmn *ch = hdr_cmn::access(p);
	pp_bytes_ += ch->size();
	pp_cnt_++;
	if(pp_cnt_ == pp_wnd_) {
		// calculate new value, reset measurement & apply new value to agents
		pp_throughput_ = (double)(pp_bytes_)/(now - pp_start_);
		pp_start_ = now;
		pp_bytes_ = 0;
		pp_cnt_ = 0;
		a_update_delta();
		A_list* pos = a_list_;
		while(pos) {
			pos->idle = 1;
			pos = pos->next;
		}
	}
}

// Calculate available BW and convert to suitable window size

double PPgauge::bw_avail() {
	double total = link_->bandwidth()/8; // total is in bytes
	double rest = total - pp_throughput_;
	return(rest);
}

double PPgauge::bw_equal(double bw, int shares) {
	double tmp = bw;
	if((int) bw == 0) tmp = -0.1*link_->bandwidth()/8;
	if(shares > 0) return(tmp/shares);
	return(tmp);
}

double PPgauge::bw2win(double bw, int seg_size) {
	double buffer = 1500*100;
	double delay = 2*link_->delay();
	double byte_wnd = bw*delay;
	double win = byte_wnd/seg_size;
	if((win >= 0) && (win <1)) win = 1;
	return(win);
}

int PPgauge::a_notIdle_cnt() {
	A_list* pos = a_list_;
	int cnt = 0;
	while(pos) {
		if(!pos->idle) cnt++;
		pos = pos->next;
	}
	a_cnt_ = cnt;
	if(cnt) return(cnt);
	return(1);
}

double PPgauge::max_share() {
	return((c_wnd_*(link_->bandwidth()/8)*2*link_->delay())/1500/a_notIdle_cnt());
}

// recv, command & trace

void PPgauge::recv(Packet* p, Handler* h) {
	hdr_ip* iph = hdr_ip::access(p);
	nsaddr_t addr = iph->daddr();
	nsaddr_t port = iph->dport();
	hdr_tcp* tcph = hdr_tcp::access(p);
	int seq = tcph->seqno();
	A_list* pos = a_list_;
	while(pos) {
		if((pos->agent->addr()==addr)&&(pos->agent->port()==port)) {
			pos->idle = 0;
			if(seq==0) {
				pos->syn=1;
				a_update_delta();
			} else {
				pos->syn=0;
			}
		}
		pos = pos->next;
	}
	pp_calc(p);
	send(p, h);
	idle_timer_.resched(timeout_);
}

int PPgauge::command(int argc, const char*const* argv) {
	Tcl& tcl = Tcl::instance();
	if(argc == 2) {
		if (strcmp(argv[1], "pp_throughput") == 0) {
			tcl.resultf("%lf",double(pp_throughput_));
			return (TCL_OK);
		} else if (strcmp(argv[1], "pp_def_wnd") == 0) {
			tcl.resultf("%d",pp_def_wnd_);
			return (TCL_OK);
		} else if (strcmp(argv[1], "test") == 0) {
			double bwa = bw_avail();
			double wina= bw2win(bwa, 1500);
			double bwe = bw_equal(0, 1);
			double wine= bw2win(bwe, 1500);
			double bwl = link_->bandwidth();
			double del = 2*link_->delay();
			tcl.resultf("Link bw: %lf, delay: %lf, Avail. BW: %lf, window: %lf 10 shares of: %lf, window: %d",
			             bwl, del, bwa, wina, bwe, (int) wine);
			return (TCL_OK);
		} else if (strcmp(argv[1], "print_agents") == 0) {
			a_print();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "attach") == 0) {
			int mode;
			const char* id = argv[2];
			channel_ = Tcl_GetChannel(tcl.interp(), (char*)id, &mode);
			if (channel_ == 0) {
				tcl.resultf("trace: can't attach %s for writing", id);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		} else if (strcmp(argv[1],"set_pp_def_wnd") == 0) {
			pp_def_wnd_ = atoi(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1],"pp_wnd_") == 0) {
			pp_wnd_ = atoi(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1],"set_c_wnd") == 0) {
			c_wnd_ = atoi(argv[2]);
			return (TCL_OK);
		} else if (strcmp(argv[1],"link") == 0) {
			link_ = (LinkDelay *)TclObject::lookup(argv[2]);
			if (link_ == 0) return TCL_ERROR;
			return (TCL_OK);
		} else if (strcmp(argv[1],"insert-agent") == 0) {
			StTcpSink* a = dynamic_cast<StTcpSink *>(TclObject::lookup(argv[2]));
			if(a) {
				a_insert(a);
				return (TCL_OK);
			}
			return (TCL_ERROR);
		} else if (strcmp(argv[1],"remove-agent") == 0) {
			StTcpSink* a = dynamic_cast<StTcpSink *>(TclObject::lookup(argv[2]));
			if(a) {
				a_remove(a);
				return (TCL_OK);
			}
			return (TCL_ERROR);
		}
	}
	return (Connector::command(argc, argv));
}

void PPgauge::trace(TracedVar* v) {
	char wrk[500];
	double now = Scheduler::instance().clock();
	// Double TracedVars
	if(!strcmp(v->name(), "pp_throughput_") || !strcmp(v->name(), "c_wnd_")) {
		double x = double(*((TracedDouble*) v));
		sprintf(wrk,"%lf %s %lf", now, v->name(), x);
	// Int TracedVars
	} else if(!strcmp(v->name(), "pp_wnd_") || !strcmp(v->name(), "a_cnt_")) {
		int y = int(*((TracedInt*) v));
		sprintf(wrk,"%lf %s %d", now, v->name(), y);
	}
	int n = strlen(wrk);
	wrk[n] = '\n';
	wrk[n+1] = 0;
	if (channel_) (void)Tcl_Write(channel_, wrk, n+1);
	wrk[n] = 0;
	return;
}

void PPgauge::a_insert(StTcpSink* a) {
	A_list* entry = new A_list;
	entry->next = a_list_;
	a_list_ = entry;
	a_list_->agent = a;
	a_list_->port  = a->port();
	a_cnt_++;
	a->rwnd_ = 1;
}

void PPgauge::a_remove(StTcpSink* a) {
	A_list* pos = a_list_;
	A_list* prev = 0;
	while((pos) && (pos->agent != a)) {
		prev = pos;
		pos = pos->next;
	}
	prev->next = pos->next;
	delete(pos);
	if(a_cnt_ > 0) a_cnt_--;
}

void PPgauge::a_remove(int port) {
	A_list* pos = a_list_;
	A_list* prev = 0;
	while((pos) && (pos->port != port)) {
		prev = pos;
		pos = pos->next;
	}
	prev->next = pos->next;
	delete(pos);
	if(a_cnt_ > 0) a_cnt_--;
}

void PPgauge::a_update_absolute() {
	double share = max_share();
	share = max_share()/a_notIdle_cnt();
	A_list* pos = a_list_;
	while(pos) {
		pos->agent->rwnd_ = (int) share;
		pos = pos->next;
	}
}

void PPgauge::a_update_delta() {
	A_list* pos = a_list_;
	double rwnd_sum = 0;
	double bw = bw_avail();
	double delta;
	double more = 0;
	while(pos) {
		if(pos->agent->rwnd_ > (int) (max_share()+1)) {
			more += pos->agent->rwnd_ - max_share();
			pos->agent->rwnd_ = (int) (max_share()+1);
		}
		if(pos->agent->rwnd_ < 1) pos->agent->rwnd_ = 1;
		if(!pos->idle) rwnd_sum += (double) pos->agent->rwnd_;
		pos = pos->next;
	}
	pos = a_list_;
	while(pos) {
		if(!pos->idle) {
			if(bw + more > 0) {
				delta = (more + bw2win(bw, 1500))*((2*rwnd_sum - (double)(pos->agent->rwnd_)*a_notIdle_cnt())/(a_notIdle_cnt()*rwnd_sum));
			} else {
				delta = (more + bw2win(bw, 1500))*(double)(pos->agent->rwnd_)/rwnd_sum;
			}
			pos->agent->rwnd_ += (int) (delta + 1);
			if(pos->agent->rwnd_ > (int) (max_share()+1))
				pos->agent->rwnd_ = (int) (max_share()+1);
			if(pos->agent->rwnd_ < 1) pos->agent->rwnd_ = 1;
		}
		pos = pos->next;
	}
}

// Na tin xanagrapsw, na epistrefei char* kai na min grafei sto stdout
void PPgauge::a_print() {
	A_list* pos = a_list_;
	printf("The agent list contains:\n");
	while(pos) {
		printf("\tAgent: %d \t Port: %d\n", pos->agent, pos->port);
		pos = pos->next;
	}
}

int PPgauge::max_rwnd() {
	A_list* pos = a_list_;
	int max = 0;
	int tmp;
	while(pos) {
		tmp = pos->agent->rwnd_;
		if(tmp > max) max = tmp;
		pos = pos->next;
	}
	return(max);
}
